from copy import copy, deepcopy


class GameStateDetector(object):
    """
    Helps keep track of game state in log messages

    Attributes:
        data (dict): game states and list of their active periods
        name (str): data identifier
    """
    def __init__(self):
        self._identifier = 'GameState_'
        self._data = {}
        self._current_state = None
        self._name = 'GameState'

    @property
    def data(self):
        return self._data

    @property
    def name(self):
        return self._name

    def check(self, message):
        """
        Checks if message contains game state information.

        :param message: single log message
        :type message: str
        :return: network_events_data
        :rtype: bool
        """
        return self._identifier in message or 'StopRecord' in message

    @staticmethod
    def parse(message):
        """
        Returns state_name extracted from message
        and any additional data the message contained
        in a list of str.

        :param message:
        :type message: str
        :return: state_name, additiona_data
        :rtype: str, list
        """
        parts = message.split()
        state_name = parts[0][10:]
        additional_data = parts[1:]

        return state_name, additional_data

    def _init_state(self, state_name, timestamp, additional_data):
        # Inserts state list into self._data
        self._data[state_name] = {}
        self._data[state_name]['timestamps'] = [[timestamp]]
        self._data[state_name]['data'] = [additional_data]

    def _check_if_state_name_previously_observed(self, state_name):
        return state_name in self._data

    def add(self, message, timestamp):
        """
        Adds information from message and its timestamp to GameStateDetector.data

        :param message:
        :type message: str
        :param timestamp:
        :type timestamp: float
        :return:
        """
        if message == 'StopRecord':
            self._data[self._current_state]['timestamps'][-1].append(timestamp)
        else:
            state_name, additional_data = GameStateDetector.parse(message)
            if self._check_if_state_name_previously_observed(state_name):
                self._data[self._current_state]['timestamps'][-1].append(timestamp)
                self._data[state_name]['timestamps'].append([timestamp])
                self._data[state_name]['data'].append(additional_data)
                self._current_state = state_name
            else:
                self._init_state(state_name, timestamp, additional_data)
                if not (self._current_state is None):
                    self._data[self._current_state]['timestamps'][-1].append(timestamp)
                self._current_state = state_name

    def finalize_message_processing(self, final_timestamp):
        """Ensures the final GameState epoch has end timestamp

        :param float final_timestamp: the end timestamp used if not already in place
        """
        if len(self._data[self._current_state]['timestamps'][-1]) == 1:
            self._data[self._current_state]['timestamps'][-1].append(final_timestamp)


class RewardDetector(object):
    """
    Helps keep track of rewards in log messages

    Attributes:
        data (dict): game states and list of their active periods
        name (str): data identifier
    """
    def __init__(self):
        self._identifiers = ('Reward pellet',
                             'Reward milk',
                             'FEEDER pellet',
                             'FEEDER milk')
        self._data = {}
        self._name = 'Reward'

    @property
    def data(self):
        return self._data

    @property
    def name(self):
        return self._name

    def check(self, message):
        """
        Checks if message contains reward information.

        :param message: single log message
        :type message: str
        :return: network_events_data
        :rtype: bool
        """
        return any([x in message for x in self._identifiers])

    @staticmethod
    def parse(message):
        """
        Extracts reward type, feeder ID and action from message.
        Action is reward quantity as str or 'inactivated' message.

        :param message:
        :type message: str
        :return: reward_type, feeder_ID, data
        :rtype: str, str, str
        """
        reward_type, feeder_id, data = message.split()[1:4]

        return reward_type, feeder_id, data

    def _check_if_reward_type_previously_observed(self, reward_type):
        return True if reward_type in self._data else False

    def _check_if_feeder_id_previously_observed(self, reward_type, feeder_id):
        return True if feeder_id in self._data[reward_type] else False

    def _create_reward_type(self, reward_type):
        self._data[reward_type] = {}

    def _create_feeder_id(self, reward_type, feeder_id):
        self._data[reward_type][feeder_id] = {'timestamps': [], 'data': []}

    def add(self, message, timestamp):
        """
        Adds information from message and its timestamp to RewardDetector.data

        :param message:
        :type message: str
        :param timestamp:
        :type timestamp: float
        :return:
        """
        reward_type, feeder_id, data = RewardDetector.parse(message)
        if not self._check_if_reward_type_previously_observed(reward_type):
            self._create_reward_type(reward_type)
        if not self._check_if_feeder_id_previously_observed(reward_type, feeder_id):
            self._create_feeder_id(reward_type, feeder_id)
        self._data[reward_type][feeder_id]['timestamps'].append(timestamp)
        self._data[reward_type][feeder_id]['data'].append(data)


class SignalDetector(object):
    """
    Helps keep track of rewards in log messages

    Attributes:
        data (dict): game states and list of their active periods
        name (str): data identifier
    """
    def __init__(self, negative_audio_signal_duration=None):
        """
        :param float negative_audio_signal_duration: required for completing NegativeAudioSignal epochs
        """

        self._identifiers = ('AudioSignal',
                             'NegativeAudioSignal',
                             'LightSignal')
        self._data = {}
        self._name = 'Signal'
        self._negative_audio_signal_duration = negative_audio_signal_duration

    @property
    def data(self):
        return self.convert_start_stop_timestamps_to_timeline()

    @property
    def name(self):
        return self._name

    @property
    def negative_audio_signal_duration(self):
        return self._negative_audio_signal_duration

    def check(self, message):
        """
        Checks if message contains reward information.

        :param message: single log message
        :type message: str
        :return: network_events_data
        :rtype: bool
        """
        return any([x in message for x in self._identifiers])

    @staticmethod
    def parse(message):
        """
        Extracts reward type, feeder ID and action from message.
        Action is reward quantity as str or 'inactivated' message.

        :param message:
        :type message: str
        :return: signal_type, data
        :rtype: str, str
        """
        signal_type, data = message.split()

        return signal_type, data

    def _check_if_signal_type_previously_observed(self, signal_type):
        return True if signal_type in self._data else False

    def _create_signal_type(self, signal_type):
        self._data[signal_type] = {'timestamps': [], 'data': []}

    def add(self, message, timestamp):
        """
        Adds information from message and its timestamp to RewardDetector.data

        :param message:
        :type message: str
        :param timestamp:
        :type timestamp: float
        :return:
        """
        signal_type, data = SignalDetector.parse(message)
        if not self._check_if_signal_type_previously_observed(signal_type):
            self._create_signal_type(signal_type)
        self._data[signal_type]['timestamps'].append(timestamp)
        self._data[signal_type]['data'].append(data)

    def convert_start_stop_timestamps_to_timeline(self):
        """
        Converts SignalDetector.data Start and Stop points into
        timeline of On periods. Works for signal_type:
        'AudioSignal' and 'LightSignal'.
        Returns a converted copy of SignalDetector.data.

        :return: data
        :rtype: dict
        """
        data = deepcopy(self._data)
        signal_candidates = ('AudioSignal', 'LightSignal')
        for signal_type in signal_candidates:
            if signal_type in data:
                epochs = []
                signal_on = False
                for timestamp, event_data in zip(data[signal_type]['timestamps'],
                                                 data[signal_type]['data']):
                    if not signal_on and event_data == 'Start':
                        epochs.append([deepcopy(timestamp)])
                        signal_on = True
                    elif signal_on and event_data == 'Stop':
                        epochs[-1].append(deepcopy(timestamp))
                        signal_on = False
                    else:
                        raise ValueError('Unexpected sequence of Start Stop for a signal')
                # Overwrite timestamps key with epochs
                data[signal_type]['timestamps'] = epochs

        # Append end timestamp to negative audio signal timestamps
        if 'NegativeAudioSignal' in data:
            if self.negative_audio_signal_duration is None:
                raise Exception(
                    'NegativeAudioSignal in messages, but negative_audio_signal_duration not provided.'
                )
            signal_type = 'NegativeAudioSignal'
            for i, timestamp in enumerate(data[signal_type]['timestamps']):
                data[signal_type]['timestamps'][i] = [timestamp, timestamp + self.negative_audio_signal_duration]

        return data


class LogParser(object):
    """
    Extracts Task related messages from list of strings
    and presents in easy to use format in class attributes.

    Attributes:
        data (dict): complete set of data detected in messages
        messages (list): task related messages
        timestamps (list): timestamps of task related messages
    """

    def __init__(self, messages, timestamps, task_settings=None, final_timestamp=None):
        """Constructor for LogParser class

        Initializes the LogParser class for a full log of a single session.

        Task relevant information is extracted and made available through query methods.

        messages   - list of str
        timestamps - list of float

        :param messages:
        :type messages: list
        :param timestamps:
        :type timestamps: list
        :param dict task_settings: output fron NWBio.load_settings()['TaskSettings']
        :param float final_timestamp: used to close any open epochs at the end of recording
        """
        # Parse input
        if not (task_settings is None) and 'NegativeAudioSignal' in task_settings:
            negative_audio_signal_duration = task_settings['NegativeAudioSignal']
        else:
            negative_audio_signal_duration = None
        # Initialize message detectors
        self.detectors = (GameStateDetector(), 
                          RewardDetector(), 
                          SignalDetector(negative_audio_signal_duration=negative_audio_signal_duration))
        # Process messages
        self._messages = []
        self._timestamps = []
        for message, timestamp in zip(messages, timestamps):
            self._process_message(message, timestamp)

        # Finalize message processing in each detector if method available
        if not (final_timestamp is None):
            for detector in self.detectors:
                if hasattr(detector, 'finalize_message_processing'):
                    detector.finalize_message_processing(final_timestamp)

    def _process_message(self, message, timestamp):
        for detector in self.detectors:
            # Check if any detector identifies the message as relevant
            if detector.check(message):
                # Add message to detector data and append to local list
                detector.add(message, timestamp)
                self._messages.append(message)
                self._timestamps.append(timestamp)
                break

    @property
    def messages(self):
        return self._messages

    @property
    def timestamps(self):
        return self._timestamps

    @property
    def data(self):
        data = {}
        for detector in self.detectors:
            data[detector.name] = detector.data
        return data

    @staticmethod
    def extract_milk_task_performance(game_state_data):
        """
        Extracts task performance from game state data.

        :param game_state_data: LogParser.data for 'GameState'
        :type game_state_data: dict
        :param float final_timestamp: used to close any open epochs at the end of recording
        :return: task_data
        :rtype: dict
        """
        # Separate milk trial game state from others
        other_states = copy(game_state_data)
        milk_state = other_states.pop('MilkTrial')
        # Create dictionary for accumulating trial information
        task_data = {'nr': [],
                     'type': [],
                     'start_timestamp': [],
                     'end_timestamp': [],
                     'outcome': [],
                     'feeder_id': []}
        # Loop through all milk trials
        trial_counter = 0
        for timestamps, data in zip(milk_state['timestamps'], milk_state['data']):
            trial_counter += 1
            task_data['nr'].append(trial_counter)
            task_data['start_timestamp'].append(timestamps[0])
            task_data['end_timestamp'].append(timestamps[1])
            # Identify feeder id
            task_data['feeder_id'].append(data[1])
            # Identify whether presentation or repeat trial
            if len(data) > 2 and (data[2] == 'presentation_trial' or data[2] == 'first_repetition'):
                task_data['type'].append('present')
            else:
                task_data['type'].append('repeat')
            # Identify trial outcome
            for other_state, value in other_states.items():
                other_state_start_timestamps = [state_times[0] for state_times in value['timestamps']]
                if timestamps[1] in other_state_start_timestamps:
                    if other_state == 'MilkReward':
                        task_data['outcome'].append('successful')
                    else:
                        other_state_data = value['data'][other_state_start_timestamps.index(timestamps[1])]
                        task_data['outcome'].append(other_state_data[0])
                    break

        return task_data



# -*- coding: utf-8 -*-
import sys
from scipy.signal import butter, lfilter, decimate
import os
import numpy as np
from PyQt5 import QtWidgets
from PyQt5.QtCore import QThread, pyqtSignal
import subprocess
import multiprocessing
import threading
from time import sleep
import psutil
from datetime import datetime
import codecs


def time_string():
    return datetime.now().strftime('%H:%M:%S')


def encode_bytes(x):
    return x if isinstance(x, bytes) else codecs.latin_1_encode(x)[0]


class CPU_availability_tracker(object):
    """
    Keeps track of available cores.
    """
    def __init__(self, n_cores):
        self._available_cores = [False for i in range(n_cores)] # Start unavailable to debug locks
        self._core_locks = [multiprocessing.Lock() for i in range(n_cores)]
        self._T__check_locks = threading.Thread(target=self._check_locks)
        self._T__check_locks.daemon = True
        self._T__check_locks.start()

    def _check_locks(self, frequency=10):
        """
        Checks locks of all unavailable cores at set interval.
        Unlocked locks indicate a core is available.

        frequency - int - how many times per second to check locks
        """
        while True:
            sleep(1.0 / frequency)
            # Check for unavailable cores
            unavailable_nrs = [i for i, state in enumerate(self._available_cores) if not state]
            # Check locks for unavailable cores
            for nr in unavailable_nrs:
                if self._core_locks[nr].acquire(block=False):
                    self._core_locks[nr].release()
                    self._available_cores[nr] = True

    def check_if_core_available(self):
        return True in self._available_cores

    def wait_for_available_core(self, check_interval=0.01):
        """
        check_interval - seconds to wait between checking for available cores.
        """
        while not self.check_if_core_available():
            sleep(check_interval)

    def use_next_available(self, block=True):
        """
        Get available core number and set it to unavailable.

        block - bool - if True, waits until core is available.
                Otherwise, if core not available, returns None

        Returns:
            nr - the number of CPU core to use, starting from 0
            lock - multiprocessing.Lock instance.
                   This must be called with lock.release()
                   for the CPU core nr to become available again.
        """
        if block:
            self.wait_for_available_core()
        elif not self.check_if_core_available():
            return None
        # Get available core number and set it to unavailable
        nr = self._available_cores.index(True)
        # Lock the core lock and set core as unavailable
        self._core_locks[nr].acquire()
        self._available_cores[nr] = False

        return nr, self._core_locks[nr]


class multiprocess(object):
    """
    This class makes it easy to use multiprocessing module in a loop.
    Use map method if memory usage is not an issue, otherwise
    use the map method as an example how to use run and results methods.
    """
    def __init__(self):
        self.CPU_availability_tracker = CPU_availability_tracker(multiprocessing.cpu_count()-1 or 1)
        self.multiprocessingManager = multiprocessing.Manager()
        self.output_list = self.multiprocessingManager.list([])
        self.output_list_Lock = multiprocessing.Lock()
        self.processor_list = []
        self.run_Lock = threading.Lock()

    @staticmethod
    def processor(output_list, output_list_Lock, list_pos, 
                  cpu_lock, f, args, kwargs):
        """
        This method called by run method in a separate process to utilize multiprocessing
        """
        # Evaluate the function with input arguments
        output = f(*args, **kwargs)
        # Update output list and active process counter
        with output_list_Lock:
            output_list[list_pos] = output
        # Release CPU core lock to inform CPU_availability_tracker
        cpu_lock.release()

    def run(self, f, args=(), kwargs=None, single_cpu_affinity=False):
        """
        This method processes the function in a separate multiprocessing.Process,
        allowing the use of multiple CPU cores.
        If all CPU cores are in use, this function blocks until a core is from a process.

        f      - function to be called in a separate process
        args   - tuple of arguments for function f
        kwargs - dictionary of keyword arguments for function f

        Optional input
            single_cpu_affinity - bool - if true, 'cpu_core_nr' keyword argument
                                  is added to kwargs or owerwritten. 
                                  The value of 'cpu_core_nr'
                                  circles through the list of available CPUs.
        """
        # If kwargs is None, create empty dict
        if kwargs is None:
            kwargs = {}
        with self.run_Lock:
            # Wait for next available cpu core count
            cpu_nr, cpu_lock = self.CPU_availability_tracker.use_next_available(block=True)
            # If requested, specify CPU to use
            if single_cpu_affinity:
                kwargs['cpu_core_nr'] = cpu_nr
            # Extend list and acquire current position
            with self.output_list_Lock:
                list_pos = len(self.output_list)
                self.output_list.append(None)
            # Start independent processor
            p = multiprocessing.Process(target=multiprocess.processor,
                                        args=(self.output_list, self.output_list_Lock, list_pos,
                                              cpu_lock, f, args, kwargs))
            p.start()
            self.processor_list.append(p)

    def results(self):
        """
        This method returns a list, where each element corresponds to output
        from functions as they are passed into run method, in the same order.
        This method blocks if run method is currently being called or
        if processors have not yet finished.
        """
        # Ensure run method is not currently running
        with self.run_Lock:
            # Ensure all processors have finished
            for p in self.processor_list:
                p.join()

        return self.output_list

    @staticmethod
    def args_kwargs_list_check(n, args_list, kwargs_list):
        # Set empty argument lists if not provided
        if args_list is None:
            args_list = [[] for x in range(n)]
        if kwargs_list is None:
            kwargs_list = [{} for x in range(n)]

        return args_list, kwargs_list

    def map(self, f, n, args_list=None, kwargs_list=None, 
            single_cpu_affinity=False, max_memory_usage=1):
        """
        This function evaluates function f for number of times specified by argument n, 
        with each set of arguments in args_list and kwargs_list.
        It outputs a list where each element is the output
        from function f for each set of arguments in args_list.
        This function is more convenient to use that run and results separately,
        but it requires loading all input arguments into memory, which is not always ideal.

        f - function to evaluate
        n - number of calls to f

        Optional input:
            args_list   - list of tuples of arguments for function f
            kwargs_list - list of dictionaries of keyword arguments for function f
            single_cpu_affinity - bool - if true, 'cpu_core_nr' keyword argument
                                  is added to kwargs or owerwritten. 
                                  The value of 'cpu_core_nr'
                                  circles through the list of available CPUs.
            max_memory_usage    - float - (0.0 - 1.0) percentage of memeory that must
                                  be available for function to process next element in list.
        """
        args_list, kwargs_list = multiprocess.args_kwargs_list_check(n, args_list, kwargs_list)
        for args, kwargs in zip(args_list, kwargs_list):
            if max_memory_usage < 1:
                proceed_when_enough_memory_available(percent=(1.0-max_memory_usage))
            self.run(f, args, kwargs, single_cpu_affinity)

        return self.results()


def proceed_when_enough_memory_available(memory_needed=None, percent=None, array_size=None, dtype=None):
    """
    This function blocks until required memory is available.
    Input can be any of the following:
        memory_needed - in bytes
        percent - percent (e.g 0.75 for 75%) of total memory to be available
        array_size and dtype - to compute memory needed to store array in memory
    """
    # Compute the memory_needed if not provided directly
    if memory_needed is None:
        if percent is None:
            if array_size is None or dtype is None:
                raise ValueError('Input needed.')
            else:
                memory_needed = array_size * dtype(1).nbytes
        else:
            memory_needed = psutil.virtual_memory().total * percent
    # Wait until this amount of memory is available
    memory_available = False
    while not memory_available:
        if psutil.virtual_memory().available > memory_needed:
            memory_available = True
        else:
            sleep(1)

    return memory_available


def lowpass_and_downsample(signal_in, sampling_rate_in, sampling_rate_out, 
                           suppress_division_by_two_error=False):
    """
    Implements `scipy.signal.decimate` method with FIR forward pass filter and
    phase shift correction.

    signal_in         - numpy array with shape (N,).
                        Input shapes (N, 1) and (1, N) are also accepted but these
                        are flattened and output shape is (N,).
                        The output will have the same dtype as signal_in.
    sampling_rate_in  - int - sampling rate of input signal
    sampling_rate_out - int - desired sampling rate of output signal.
                        The division of sampling_rate_in by sampling_rate_out must have 
                        zero remainder.
                        Furthermore, the resulting quotient must have 0 remainder when
                        divided by 2. Otherwise scipy package gives a warning about
                        bad coefficients.

    Optional:

    suppress_division_by_two_error - bool - allows using sampling_rate_out that does not
                                            satisfy the requirement that the resulting 
                                            quotient with sampling_rate_in has remainder 0 
                                            when divided by sampling_rate_in. 
                                            Default is False.

    """

    # Ensure input signal has the correct shape
    if len(signal_in.shape) > 1:
        if len(signal_in.shape) == 2 and signal_in.shape[1] == 1:
            signal_in = signal_in.squeeze()
        elif len(signal_in.shape) == 2 and signal_in.shape[0] == 1:
            signal_in = signal_in.squeeze()
        else:
            raise Exception('signal_in must have shape (N,), (N, 1) or (1, N)')

    # Ensure sampling rates can be divided without remainder
    if sampling_rate_in % sampling_rate_out != 0:
        raise Exception('sampling_rate_out must be a factor of sampling_rate_in.')

    # Compute downsampling factor
    downsampling_factor = int(sampling_rate_in / sampling_rate_out)

    # Verify that downsampling factor can be divided by 2 without remainder
    if not suppress_division_by_two_error:
        if downsampling_factor % 2 != 0:
            raise Exception('Quotient of sampling_rate_in and sampling_rate_out '
                            + 'must have zero remainder when divided by 2.')

    # Filter and downsample the signal
    signal_out = decimate(signal_in, downsampling_factor, ftype='fir', zero_phase=True)

    # Ensure output is in same dtype as input signal
    if not (signal_in.dtype is signal_out.dtype):
        signal_out = signal_out.astype(signal_in.dtype)

    return signal_out


def butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a

def butter_bandpass_filter(signal_in, sampling_rate=30000.0, highpass_frequency=300.0, lowpass_frequency=6000.0, filt_order=4):
    b, a = butter_bandpass(highpass_frequency, lowpass_frequency, sampling_rate, order=filt_order)
    signal_out = lfilter(b, a, signal_in)
    return signal_out


def butter_lowpass(cutoff, fs, order=5):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return b, a


def butter_lowpass_filter(signal_in, lowpass_frequency=125.0, sampling_rate=30000.0, filt_order=4):
    if len(signal_in.shape) > 1:
        raise Exception('signal_in must have shape (N,), but has shape ' + str(signal_in.shape))
    b, a = butter_lowpass(lowpass_frequency, sampling_rate, order=filt_order)
    signal_out = lfilter(b, a, signal_in)
    return signal_out


# Print iterations progress
def print_progress(iteration, total, prefix='', suffix='', decimals=1, bar_length=40, initiation=False):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """
    if not initiation:
        sys.stdout.flush()
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '█' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write('%s |%s| %s%s %s\r' % (prefix, bar, percents, '%', suffix)),

    if iteration >= total:
        sys.stdout.write('\n')
    sys.stdout.flush()
# # 
# # Sample Usage
# # 
# from time import sleep
# # A List of Items
# items = list(range(0, 57))
# l = len(items)
# # Initial call to print 0% progress
# print_progress(0, l, prefix = 'Progress:', suffix = 'Complete',initiation=True)
# for i, item in enumerate(items):
#     # Do stuff...
#     sleep(0.1)
#     # Update Progress Bar
#     print_progress(i + 1, l, prefix = 'Progress:', suffix = 'Complete')


def get_position_data_edges(fpath):
    # Get position data first and last timestamps
    PosFilePath = os.path.join(fpath,'PosLogComb.csv')
    pos_csv = np.genfromtxt(PosFilePath, delimiter=',')
    pos_timestamps = np.array(pos_csv[:,0], dtype=np.float64)
    pos_edges = [pos_timestamps[0], pos_timestamps[-1]]

    return pos_edges


def channels_tetrode(nchan):
    # Returns the tetrode number this channel is on (all counting starts from 0)
    nchan = float(nchan + 1)
    ntet = np.ceil(nchan / 4) - 1
    ntet = int(ntet)

    return ntet


def tetrode_channels(ntet):
    # Returns the list of channels on this tetrode all counting starts from 0)
    nchan = ntet * 4
    nchans = [nchan, nchan + 1, nchan + 2, nchan + 3]

    return nchans

def get_tetrode_nrs(channels):
    firstTet = channels_tetrode(channels[0])
    lastTet = channels_tetrode(channels[-1])
    tetrode_nrs = list(range(firstTet, lastTet + 1, 1))

    return tetrode_nrs

def import_subdirectory_module(subdirectory, module_name):
    path = list(sys.path)
    sys.path.insert(0, subdirectory)
    try:
        module = __import__(module_name)
    finally:
        sys.path[:] = path

    return module

def test_pinging_address(address='localhost'):
    """
    Returns True if successful in pinging the input IP address, False otherwise
    """
    with open(os.devnull, 'w') as devnull:
        result = subprocess.call(['ping', '-c', '3', address], stdout=devnull, stderr=devnull)
    if result == 0:
        return True
    else:
        return False


def closest_argmin(a, b):
    """Returns the closest matching element indices in b for each element in a.

    :param numpy.ndarray a: shape (N,)
    :param numpy.ndarray b: shape (M,)
    :return: closest_b_indices
    :rtype: numpy.ndarray
    """
    b_length = b.size
    sorted_idx_b = b.argsort()
    sorted_b = b[sorted_idx_b]
    sorted_idx = np.searchsorted(sorted_b, a)
    sorted_idx[sorted_idx == b_length] = b_length - 1
    mask = (sorted_idx > 0) & (np.abs(a - sorted_b[sorted_idx-1]) < np.abs(a - sorted_b[sorted_idx]))

    return sorted_idx_b[sorted_idx-mask]


def openSingleFileDialog(load_save, directory_path=os.path.expanduser("~"), suffix='', caption='Choose File'):
    """Opens a GUI dialog for browsing files when closed returns full path to file.

    If no file is selected but dialog is closed, then this function returns None.

    :param str load_save: either 'save' or 'load', which determines if all or only existing files can be selected
    :param str directory_path: optionally specify starting directory
    :param str suffix: specify filename suffixes to filter for
    :param str caption: caption to display on dialog
    :return: fpath
    """
    # Pops up a GUI to select a single file.
    dialog = QtWidgets.QFileDialog(directory=directory_path, caption=caption)
    if load_save is 'save':
        dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
    elif load_save is 'load':
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFile)
    dialog.setViewMode(QtWidgets.QFileDialog.List) # or Detail
    if len(suffix) > 0:
        dialog.setNameFilter('*.' + suffix)
        dialog.setDefaultSuffix(suffix)
    if dialog.exec_():
        # Get path and file name of selection
        tmp = dialog.selectedFiles()
        selected_file = str(tmp[0])
        return selected_file


def show_message(message, message_more=None):
    # This function is used to display a message in a separate window
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Information)
    msg.setText(message)
    if message_more:
        msg.setInformativeText(message_more)
    msg.setWindowTitle('Message')
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    return msg.exec()


class QThread_with_completion_callback(QThread):
    finishedSignal = pyqtSignal()
    """
    Allows calling funcitons in separate threads with a callback function executed at completion.

    Note! Make sure QThread_with_completion_callback instance does not go out of scope during execution.
    """
    def __init__(self, callback_function, function, return_output=False, callback_args=(), function_args=()):
        """
        The callback_function is called when function has finished.
        The function is called with any input arguments that follow it.
        """
        super(QThread_with_completion_callback, self).__init__()
        if return_output:
            self.finishedSignal.connect(lambda: callback_function(self.function_output, *self.callback_args))
        else:
            self.finishedSignal.connect(lambda: callback_function(*self.callback_args))
        self.function = function
        self.callback_args = callback_args
        self.function_args = function_args
        self.start()

    def run(self):
        self.function_output = self.function(*self.function_args)
        self.finishedSignal.emit()


def clearLayout(layout, keep=0):
    # This function clears the layout so that it could be regenerated
    if layout is not None:
        while layout.count() > keep:
            item = layout.takeAt(keep)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            else:
                clearLayout(item.layout())


class ClassInSeparateProcess(object):
    """
    This class can be used to instantiate and control another class in a separate process.

    The controlled class must have a `closed` attribute that indicates
    if the class has been closed after instantiation.

    Inputs to the task must be functional in a separate process created with multiprocessing package.
    """

    def __init__(self, cls, args=None, kwargs=None):
        """
        :param cls: class with `close` attribute to indicate if class has been closed after instantiation
        :param tuple args: positional arguments to pass into class
        :param dict kwargs: keyword arguments to pass into class
        """

        command_receiver_pipe, self._command_sender_pipe = multiprocessing.Pipe()

        self._class_process = multiprocessing.Process(target=self.class_process_method,
                                                      args=(cls, command_receiver_pipe,
                                                            () if args is None else args,
                                                            {} if kwargs is None else kwargs))
        self._class_process.start()

        while True:
            if self._command_sender_pipe.poll(0.1):
                if self._command_sender_pipe.recv() == 'initialization successful':
                    print('Class {} started successfully in separate process.'.format(cls))
                    break
                else:
                    raise Exception('Unknown input from class {} process.'.format(cls))

    @property
    def class_process_active(self):
        """Returns True if the processes with class instance is active and False otherwise.

        :return: alive_bool
        :rtype: bool
        """
        return self._class_process.is_alive()

    @staticmethod
    def class_process_method(cls, command_receiver_pipe, args, kwargs):
        """This method initializes the class and runs until it has completed `close` command.

        Any string received via `command_receiver_pipe` are executed as class method names
        and output returned via `command_receiver_pipe`.

        :param cls: class with `close` attribute to indicate if class has been closed after instantiation
        :param multiprocessing.connection command_receiver_pipe:
        :param tuple args: positional arguments to pass into class
        :param dict kwargs: keyword arguments to pass into class
        """

        class_instance = cls(*args, **kwargs)

        command_receiver_pipe.send('initialization successful')

        while not class_instance.closed:

            if command_receiver_pipe.poll(0.1):

                method_name = command_receiver_pipe.recv()

                ret = getattr(class_instance, method_name)()

                command_receiver_pipe.send(ret)

    def call_class_method(self, name):
        """Calls the named method on the class running in another process and returns output.

        This method blocks until the method in another process returns.

        :param str name: name of the method to be called
        :return: return_from_method
        """
        if self.class_process_active:
            self._command_sender_pipe.send(name)
            return self._command_sender_pipe.recv()
        else:
            raise Exception('Class Process is not running. Can not receive commands.')



### This script produces a window where it plots current tracking position
### and also history of tracked position.

### By Sander Tanni, January 2018, UCL

import sys
import pyqtgraph as pg
from PyQt5 import QtWidgets, QtGui
import numpy as np
from PyQt5.QtCore import QTimer
from time import sleep, time
from scipy import ndimage
from copy import copy, deepcopy


def angle_clockwise(p1, p2, invertedY=True):
    # Takes two numpy 2D vectors and computes the angle of their difference
    # In other words, computes the angle of a vector from p1 to p2
    p0 = np.array(p2) - np.array(p1)
    angle_rad = np.arctan2(p0[0], p0[1])
    angle_rad = np.arctan2(np.sin(angle_rad), np.cos(angle_rad))
    angle_deg = np.degrees(angle_rad)
    # Correct angle for inveted image
    if invertedY:
        angle_deg = -angle_deg + 180

    return angle_deg


class PosPlot(QtWidgets.QWidget):
    """
    Displays position data as provided by OnlineTracker class.

    Closes automatically when OnlineTracker class is no longer active.
    """

    def __init__(self, processed_position_list, position_histogram_array, position_histogram_dict,
                 online_tracker_is_alive, arena_size, LED_angle=None):
        super().__init__()

        self.processed_position_list = processed_position_list
        self.position_histogram_array = position_histogram_array
        self.online_tracker_is_alive = online_tracker_is_alive
        self.arena_size = arena_size
        self.LED_angle = LED_angle

        self.histogram_array_shape = position_histogram_dict['array_shape']
        self.histogramParameters = position_histogram_dict['parameters']

        # Only continue once first two position samples are available
        while len(self.processed_position_list) < 2:
            sleep(0.1)

        # Initialize plot window
        self.PlotGraphicsWidget = pg.GraphicsLayoutWidget()
        self.plotBox = self.PlotGraphicsWidget.addViewBox(enableMouse=False)
        self.plotBox.setAspectLocked(True)
        self.imageItem = pg.ImageItem()
        self.plotBox.addItem(self.imageItem)
        self.plotBox.invertY(True)

        # Create color map box
        self.ColormapGraphicsWidget = pg.GraphicsLayoutWidget()
        self.ColormapGraphicsWidget.setFixedWidth(50)
        self.colorMapBox = self.ColormapGraphicsWidget.addViewBox(enableMouse=False)
        self.colorMapBox.setAspectLocked(False)
        self.colorMapImageItem = pg.ImageItem()
        self.colorMapBox.addItem(self.colorMapImageItem)

        # Create UI controls
        self.maxValueBox = QtWidgets.QLineEdit('10')
        self.thresholdValueBox = QtWidgets.QLineEdit('2')
        self.maxValueBox.returnPressed.connect(self.prepareColormap)
        self.thresholdValueBox.returnPressed.connect(self.prepareColormap)
        self.smoothingValueBox = QtWidgets.QLineEdit('3')
        self.smoothingValueBox.returnPressed.connect(self.updateSmoothingValue)
        self.updateSmoothingValue()
        self.showPathButton = QtWidgets.QPushButton('Show Path')
        self.showPathButton.setCheckable(True)
        self.showPathButton.setChecked(False)
        self.showPathButton.clicked.connect(self.showPath)

        # Put all plots into main window and display
        self.setWindowTitle('Cumulative Position Plot')
        XandYratio = (float(self.arena_size[0])
                      / float(self.arena_size[1]))
        self.resize(int(700 * XandYratio) + 200, 700)
        vboxWidget = QtWidgets.QWidget()
        vboxWidget.setFixedWidth(100)
        vbox = QtWidgets.QVBoxLayout(vboxWidget)
        vbox.addWidget(QtWidgets.QLabel('Maximum'))
        vbox.addWidget(QtWidgets.QLabel('value'))
        vbox.addWidget(QtWidgets.QLabel('n visits'))
        vbox.addWidget(self.maxValueBox)
        vbox.addStretch()
        vbox.addWidget(QtWidgets.QLabel('Threshold'))
        vbox.addWidget(QtWidgets.QLabel('value'))
        vbox.addWidget(self.thresholdValueBox)
        vbox.addStretch()
        vbox.addWidget(QtWidgets.QLabel('Smoothing'))
        vbox.addWidget(QtWidgets.QLabel('in cm'))
        vbox.addWidget(self.smoothingValueBox)
        vbox.addWidget(self.showPathButton)
        hbox = QtWidgets.QHBoxLayout(self)
        hbox.addWidget(self.PlotGraphicsWidget)
        hbox.addWidget(self.ColormapGraphicsWidget)
        hbox.addWidget(vboxWidget)
        self.show()

        # Prepare plot
        self.prepareColormap()
        self.updatePlotAxes()
        self.draw_arena_boundaries()
        self.arrow = pg.ArrowItem(pos=(0, 0), angle=0, headLen=0, tailLen=0, headWidth=0, tailWidth=0)
        self.plotBox.addItem(self.arrow)

        # Start constant update of the plot
        self.keep_updating_plot = True
        plotUpdateInterval = 100  # This sets the plot update interval in milliseconds
        self.cumulativePlot_timer = QTimer()
        self.cumulativePlot_timer.timeout.connect(self.updatePlot)
        self.cumulativePlot_timer.start(plotUpdateInterval)

        # Keep track of whether Online Tracker is still alive. If not, close position plot
        self.online_tracking_data_alive_checker_timer = QTimer()
        self.online_tracking_data_alive_checker_timer.timeout.connect(self.close_if_online_tracker_not_alive)
        self.online_tracking_data_alive_checker_timer.start(100)

    def close_if_online_tracker_not_alive(self):
        if not self.keep_updating_plot or not self.online_tracker_is_alive.get():
            self.close()

    def updatePlotAxes(self):
        binSize = self.histogramParameters['binSize']
        margins = self.histogramParameters['margins']
        xRange = (0, (self.arena_size[0] + 2 * margins) / binSize)
        yRange = (0, (self.arena_size[1] + 2 * margins) / binSize)
        self.plotBox.setRange(xRange=xRange, yRange=yRange)

    def prepareColormap(self):
        # This function prepares colormap for displaying the histogram
        # Get colormap parameters
        self.hist_max_range = float(str(self.maxValueBox.text()))
        self.hist_threshold_value = float(str(self.thresholdValueBox.text()))
        # Set colors for below and above threshold
        # Change colorval here to edit colors
        colormidpos = self.hist_threshold_value / float(self.hist_max_range)
        colorpos = np.array([0.0, 0.001, colormidpos, colormidpos + 0.001, 1.0], dtype=np.float32)
        colorval = np.array([[0,0,0], [255, 0, 0], [127,127,255], [0,255,0], [0,255,255]], dtype=np.ubyte)
        # Interpolate at high resolution to provide input to pyqtgraph colormap function
        colorposFull = np.array([], dtype=np.float32)
        colorvalFull = np.zeros((0,3), dtype=np.float32)
        cincr = 0.0001
        for cstep in range(len(colorpos) -  1):
            cpos1, cpos2 = colorpos[cstep:cstep+2]
            tmpcolorpos = np.arange(cpos1, cpos2, cincr)
            colorposFull = np.append(colorposFull, tmpcolorpos)
            tmpcolorval = np.zeros((tmpcolorpos.size, colorval.shape[1]), dtype=np.float32)
            for ncolor in range(colorval.shape[1]):
                cval1 = colorval[cstep][ncolor]
                cval2 = colorval[cstep + 1][ncolor]
                tmpcolorval[:, ncolor] = np.interp(tmpcolorpos,
                                                   np.array([cpos1, cpos2], dtype=np.float32),
                                                   np.array([cval1, cval2], dtype=np.float32))
            colorvalFull = np.concatenate((colorvalFull, tmpcolorval.astype(np.uint8)), axis=0)
        # Form colormap lookup table for plotting
        CMap = pg.ColorMap(colorposFull, colorvalFull)
        self.CMapLut = CMap.getLookupTable(start=0.0, stop=1.0, nPts=512, alpha=False, mode='float')
        # Display colormap in a separate plot
        colormap_image = np.repeat(np.linspace(0.0,1.0,300), 15).reshape((300, 15))
        colormap_image = np.uint8(colormap_image * 255)
        self.colorMapImageItem.setImage(colormap_image.T, autoLevels=False, lut=self.CMapLut)

    def draw_arena_boundaries(self):
        # Draw boundaries of the arena to the plot
        boundaries = np.array([[0, 0],
                               [self.arena_size[0], 0],
                               [self.arena_size[0],
                                self.arena_size[1]],
                               [0, self.arena_size[1]],
                               [0, 0]])
        boundaries = boundaries + self.histogramParameters['margins']
        boundaries = boundaries / float(self.histogramParameters['binSize'])
        self.boundaryBox = pg.PlotDataItem()
        self.boundaryBox.setData(boundaries)
        self.boundaryBox.setPen(pg.mkPen('b', width=4))
        self.plotBox.addItem(self.boundaryBox)

    def updateSmoothingValue(self):
        # This is in centimeters
        self.smoothingValue = float(str(self.smoothingValueBox.text()))

    def showPath(self):
        if self.showPathButton.isChecked():
            posHistory = copy(self.processed_position_list[0:])
            posHistory = [i for i in posHistory if i is not None]
            posHistory = np.array(posHistory)[:, :2]
            posHistory = posHistory + self.histogramParameters['margins']
            posHistory = posHistory / float(self.histogramParameters['binSize'])
            self.trackedPath = pg.PlotDataItem()
            self.trackedPath.setData(posHistory)
            self.trackedPath.setPen(pg.mkPen('w', width=1))
            self.plotBox.addItem(self.trackedPath)
        else:
            self.plotBox.removeItem(self.trackedPath)

    def updatePlot(self):
        if self.keep_updating_plot and self.online_tracker_is_alive.get():

            # Get latest position data
            pastPos, currPos = self.processed_position_list[-2:]
            positionHistogram = np.reshape(self.position_histogram_array,
                                           self.histogram_array_shape)

            # Smooth histogram
            smoothGaussStd = self.smoothingValue / float(self.histogramParameters['binSize'])
            image = ndimage.gaussian_filter(np.float32(positionHistogram.T),
                                            sigma=(smoothGaussStd, smoothGaussStd),
                                            order=0)
            # Set image data maximum value as requested
            image[image > self.hist_max_range] = np.float32(self.hist_max_range)
            # Display image of histogram
            image = np.uint8(image / np.float32(self.hist_max_range) * 255)
            self.imageItem.setImage(image, autoLevels=False, lut=self.CMapLut)
            # Draw arrow for position if position data available
            self.plotBox.removeItem(self.arrow)  # Remove previous arrow
            if not (currPos is None) and not (pastPos is None):
                arrowPos = ((currPos[0] + self.histogramParameters['margins']) / self.histogramParameters['binSize'],
                            (currPos[1] + self.histogramParameters['margins']) / self.histogramParameters['binSize'])
                if not (self.LED_angle is None) and not np.any(np.isnan(currPos[2:])):
                    # Compute arrow angle with to align the line connecting the points
                    head_angle = angle_clockwise(currPos[:2], currPos[2:]) + 90 + self.LED_angle
                    # Draw the arro with the head/tip at the location of primary LED
                    self.arrow = pg.ArrowItem(pos=arrowPos, 
                                              angle=head_angle, headLen=40, tailLen=20, 
                                              headWidth=40, tailWidth=7, 
                                              brush=pg.mkBrush('b'), pen=pg.mkPen('c', width=3))
                    self.plotBox.addItem(self.arrow)
                else:
                    # If no 2nd LED data, draw arrow based on movement direction
                    head_angle = angle_clockwise(pastPos[:2], currPos[:2]) + 90
                    self.arrow = pg.ArrowItem(pos=arrowPos, 
                                              angle=head_angle, headLen=40, tailLen=0, 
                                              headWidth=40, tailWidth=7, 
                                              brush=pg.mkBrush('b'), pen=pg.mkPen('c', width=3))
                    self.plotBox.addItem(self.arrow)
            else:
                # Add invisible arrow item to keep loop consistent
                self.arrow = pg.ArrowItem(pos=[0, 0], 
                                          angle=0, headLen=0, tailLen=0, 
                                          headWidth=0, tailWidth=0, 
                                          brush=pg.mkBrush('b'))
                self.plotBox.addItem(self.arrow)

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        self.close()
        super(PosPlot, self).closeEvent(a0)

    def close(self):
        if self.keep_updating_plot:
            # Close the update loop, RPi Position tracking loop and application window
            self.keep_updating_plot = False
            self.cumulativePlot_timer.stop()
            self.online_tracking_data_alive_checker_timer.stop()
            sleep(0.25)
            print('Stopped Cumulative Position Plot.')
        super(PosPlot, self).close()



import cv2
import os
import argparse

class save_video(object):
    def __init__(self, filename, frame_width, frame_height, framerate):
        # fourcc = cv2.VideoWriter_fourcc(*'MP4V')
        self.out = cv2.VideoWriter(filename, 0x7634706d, framerate, (frame_width,frame_height))

    def write_frame(self, frame):
        self.out.write(frame)

    def close(self):
        self.out.release()

def convert(folder_path, outfile_path, framerate=25):
    '''
    Sorts files by numbers preceding .jpg extension and writes them into an mp4 file.
    '''
    # Get all files in folder
    filenames = os.listdir(folder_path)
    # Sort file names based on numbers
    filenrs = [int(x[:-4]) for x in filenames]
    filenames = [x for _,x in sorted(zip(filenrs,filenames))]
    filenrs = [int(x[:-4]) for x in filenames]
    # Load sample frame
    sampleFrame = cv2.imread(os.path.join(folder_path, filenames[0]))
    frame_height, frame_width, _ = sampleFrame.shape
    # Initialize save_video
    SV = save_video(outfile_path, frame_width, frame_height, framerate)
    # Loop through jpg files and write video
    for n, filename in enumerate(filenames):
        frame = cv2.imread(os.path.join(folder_path, filename))
        SV.write_frame(frame)
        if n % 1000 == 0:
            print(str(n) + ' frames written.')
    # Close video writer
    SV.close()

if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Convert all jpg images in folder to mp4 video.')
    parser.add_argument('folder_path', type=str, nargs=1, 
                        help='Path to folder with jpg files.')
    parser.add_argument('outfile_path', type=str, nargs=1, 
                        help='Path of video file.')
    args = parser.parse_args()
    convert(args.folder_path[0], args.outfile_path[0])
